indexing

   description: "portable child processes for POSIX."

   author: "Berend de Boer"
   date: "$Date: 2001/12/27 $"
   revision: "$Revision: #1 $"


class

   EPX_EXEC_PROCESS


inherit

   ABSTRACT_EXEC_PROCESS
      redefine
         fd_stdin,
         fd_stdout,
         fd_stderr
      end

   POSIX_FORK_ROOT
      rename -- these are only applicable if you're the child
         stdin as child_stdin,
         stdout as child_stdout,
         stderr as child_stderr,
         fd_stdin as child_fd_stin,
         fd_stdout as child_fd_stdout,
         fd_stderr as child_fd_sterr
      redefine
         wait_for
      end


creation

   make,
   make_capture_input,
   make_capture_output,
   make_capture_io,
   make_capture_all


feature -- i/o capturing

   stdin: POSIX_TEXT_FILE

   stdout: POSIX_TEXT_FILE

   stderr: POSIX_TEXT_FILE

   fd_stdin: POSIX_FILE_DESCRIPTOR

   fd_stdout: POSIX_FILE_DESCRIPTOR

   fd_stderr: POSIX_FILE_DESCRIPTOR


feature -- execute

   execute is
         -- Executes `program_name'
         -- don't forget to `wait' for this process to terminate
      local
         input_pipe,
         output_pipe,
         error_pipe: POSIX_PIPE
         in_child: BOOLEAN
      do
         if capture_input then
            create input_pipe.make
         else
            fd_stdin := Void
         end
         if capture_output then
            create output_pipe.make
         else
            fd_stdout := Void
         end
         if capture_error then
            create error_pipe.make
         else
            fd_stderr := Void
         end

         my_pid := posix_fork
         if my_pid = -1 then
            raise_posix_error
         else
            running := True
            in_child := my_pid = 0
            if in_child  then

               -- this child process has now current process characteristics
               -- open the standard io files if needed
               if capture_input then
                  create fd_stdin.attach_to_fd (STDIN_FILENO)
               end
               if capture_output then
                  create fd_stdout.attach_to_fd (STDOUT_FILENO)
               end
               if capture_error then
                  create fd_stderr.attach_to_fd (STDERR_FILENO)
               end

               -- close a captured standard file and attach it to the pipe
               if capture_input then
                  fd_stdin.make_as_duplicate (input_pipe.fdin)
                  check
                     fd_stdin.value = STDIN_FILENO
                  end
               end
               if capture_output then
                  fd_stdout.make_as_duplicate (output_pipe.fdout)
                  check
                     fd_stdout.value = STDOUT_FILENO
                  end
               end
               if capture_error then
                  fd_stderr.make_as_duplicate (error_pipe.fdout)
                  check
                     fd_stderr.value = STDERR_FILENO
                  end
               end

               -- close left over file descriptors
               if capture_input then
                  input_pipe.close
               end
               if capture_output then
                  output_pipe.close
               end
               if capture_error then
                  error_pipe.close
               end

               do_execute

            else -- this is parent

               -- make the captured i/o available as the stdout
               -- of this object.
               -- file descriptor will remain owner, so make sure not
               -- to close both stream and file descriptor
               if capture_input then
                  fd_stdin := input_pipe.fdout
                  create stdin.make_from_file_descriptor (fd_stdin, "w")
                  input_pipe.fdin.close
                  fd_stdin.make_owner
               end
               if capture_output then
                  fd_stdout := output_pipe.fdin
                  create stdout.make_from_file_descriptor (fd_stdout, "r")
                  output_pipe.fdout.close
                  fd_stdout.make_owner
               end
               if capture_error then
                  fd_stderr := error_pipe.fdin
                  create stderr.make_from_file_descriptor (fd_stderr, "r")
                  error_pipe.fdout.close
                  fd_stderr.make_owner
               end

            end
         end
      rescue
         if in_child then
            handle_execute_failure
         end
      end


feature -- actions that parent may execute

   wait_for (suspend: BOOLEAN) is
      do
         precursor (suspend)
         if is_terminated then
            -- avoid file handle leaks. We have to be careful because
            -- the streams can be closed too. And because the file
            -- descriptor also is an owner, we should avoid closing an
            -- already closed descriptor.
            if fd_stdin /= Void then
               if fd_stdin.is_open and then stdin.is_open then
                  fd_stdin.close
                  stdin.unattach
               else
                  fd_stdin.unattach
               end
            end
            if fd_stdout /= Void then
               if fd_stdout.is_open and then stdout.is_open then
                  fd_stdout.close
                  stdout.unattach
               else
                  fd_stdout.unattach
               end
            end
            if fd_stderr /= Void then
               if fd_stderr.is_open and then stderr.is_open then
                  fd_stderr.close
                  stderr.unattach
               else
                  fd_stderr.unattach
               end
            end
         end
      end


feature {NONE}

   do_execute is
      -- actually execute `program_name'
      local
         argv: ARRAY[STRING]
         cargv: ARRAY[POINTER]
         i, j: INTEGER
         r: INTEGER
      do
         create argv.make (0, arguments.count+1)
         argv.put (program_name, 0)
         from
            i := arguments.lower
            j := 1
         until
            i > arguments.upper
         loop
            argv.put (arguments.item(i), j)
            i := i + 1
            j := j + 1
         end
         cargv := ah.string_array_to_pointer_array (argv)
         r := posix_execvp (sh.string_to_pointer (program_name),
                            ah.pointer_array_to_pointer (cargv))

         -- we should never return here
         -- but if we do, r = -1
         -- And we're already forked, so we're really in a separate process
         -- There's no way to signal the parent (except through SIGCHILD)
         handle_execute_failure
      end

   handle_execute_failure is
         -- attempt to clean up, and exit gracefully.
      do
         -- not sure if creating stuff is a good idea here
         -- perhaps should be done earlier?
         -- Anyway, I want to write something to stderr
         -- Format of error message is from perl5
         child_stderr.puts ("Can't exec %"")
         child_stderr.puts (program_name)
         child_stderr.puts ("%": ")
         child_stderr.puts (errno.message)
         child_stderr.puts (".%N")
         if capture_input then
            fd_stdin.close
         end
         if capture_output then
            fd_stdout.close
         end
         if capture_error then
            fd_stderr.close
         end
         -- a failed child should not flush parent files
         minimal_exit (EXIT_FAILURE)
      rescue
         minimal_exit (EXIT_FAILURE)
      end


end
